#ifndef DSEL
/* SMCHSMode -- Serial High Speed Mode enabler V1.2b */
/* For SMC 37C665IR/666IR/669/669FR super I/O chip */
/*         37C67x/68x Enhanced super I/O chip */
/*         37C93x/93xAPM/93xFR/957FR ultra I/O chip */
/* Written by imaizumi@nisiq.net */
/* Special thanks for mio-h@ca2.so-net.or.jp */
/* Modified for Linux by mizuhara@ca2.so-net.or.jp */
/* compiler DOS:MSC6.00A, Linux:gcc */
/*
Release note:
  1.2b(25-Dec-96)
    Added support for 67x,68x(Not tested yet.)

  1.2a(01-Nov-96)
    Add 'B'(oth) port option.
  1.2(31-Oct-96)
    Add support for 669FR,93xFR,93xAPM,957FR(Not tested yet.)
    Display port address for convenience.
  1.0f
    add debug code & adaptation for linux.
  1.0d, 1.0e
    experimental,not released.
  1.0b -> 1.0c
    display string "com" -> "port", function exactly same.

Known limitation for linux:
    Perhaps, this version fail to detect/set operation rare.
    You must check result.
    In hsmode,you can use ordinaly setserial command to over 115kbps.
        230kbps: setserial <dev> spd_cust divisor 0x8002
        460kbps: setserial <dev> spd_cust divisor 0x8001
*/
#else
/* SMCDSel -- Floppy Density Select V1.2b */
/* Only for Japanese 3 mode 3.5'FDD(720K/1.2M/1.44M), Like FD235HG */
/* For SMC 37C665GT/666GT/665IR/666IR/669/669FR super I/O chip */
/*         37C67x/68x Enhanced super I/O chip */
/*         37C93x/93xAPM/93xFR/957FR ultra I/O chip */
/* Written by imaizumi@nisiq.net */
/* compiler DOS:MSC6.00A */
/* CL /Ox /DDSEL /FeSMCDSEL SMCHSMOD.C */
/*
Release note:
  1.2b(25-Jan-96)
    Source code contained to smchsmod.c.
*/
#endif

#include <stdio.h>
#include <stdlib.h>
#if defined(unix)
void port_init(void);
void port_finish(void);
#define _disable() port_init()
#define _enable() port_finish()
#if defined(linux)
int inp(unsigned);
void outp(unsigned, unsigned char);
#elif defined(__FreeBSD__)
#include <machine/cpufunc.h>
#define inp(port)   inb(port)
#define outp(port, val) outb(port, val)
#endif /* !defined(linux) && defined(__FreeBSD__) */
#else /* !defined(unix) */
#include <conio.h>
#include <dos.h>
#endif /* !defined(unix) */
#include <string.h>

void usage(void);
void srdump(int n,int start,int end);

typedef struct idn {
unsigned port;
unsigned char ini;
unsigned char idx;
unsigned char id;
unsigned char rev;
int type;
char *name;
} idn_t;

idn_t idnt[] = {
    0x3f0,0x55,0x0d,0x65,0x02,1,"665GT",    /* 0:665GT */
    0x3f0,0x55,0x0d,0x65,0x82,2,"665IR",    /* 1:665IR */
    0x3f0,0x44,0x0d,0x66,0x02,1,"666GT",    /* 2:666GT */
    0x3f0,0x44,0x0d,0x66,0x82,2,"666IR",    /* 3:666IR */
    0x3f0,0x55,0x0d,0x03,0x00,3,"669"  ,    /* 4:669 */
    0x370,0x55,0x0d,0x03,0x00,3,"669"  ,    /* 5:669 at 0x370 */
    0x3f0,0x55,0x0d,0x04,0x00,3,"669FR",    /* 6:669FR */
    0x370,0x55,0x0d,0x04,0x00,3,"669FR",    /* 7:669FR at 0x370 */
    0x3f0,0x55,0x20,0x40,0x01,4,"67x",      /* 8:67x */
    0x370,0x55,0x20,0x40,0x01,4,"67x",      /* 9:67x at 0x370 */
    0x3f0,0x55,0x20,0x48,0x01,4,"68x",      /*10:68x */
    0x370,0x55,0x20,0x48,0x01,4,"68x",      /*11:68x at 0x370 */
    0x3f0,0x55,0x20,0x02,0x01,4,"93x"  ,    /*12:93x */
    0x370,0x55,0x20,0x02,0x01,4,"93x"  ,    /*13:93x at 0x370 */
    0x3f0,0x55,0x20,0x30,0x01,4,"93xAPM",   /*14:93xAPM */
    0x370,0x55,0x20,0x30,0x01,4,"93xAPM",   /*15:93xAPM at 0x370 */
    0x3f0,0x55,0x20,0x03,0x01,4,"93xFR",    /*16:93xFR */
    0x370,0x55,0x20,0x03,0x01,4,"93xFR",    /*17:93xFR at 0x370 */
    0x3f0,0x55,0x20,0x07,0x01,4,"957FR",    /*18:957FR */
    0x370,0x55,0x20,0x07,0x01,4,"957FR",    /*19:957FR at 0x370 */
    0,0,0,0,0,0,"Other"
};

void main(int argc, char *argv[])
{
#ifndef DSEL
    char opr;
    unsigned char id,rev,r,v,cr1,cr2;
    int n,portNo,portAdr[3],i,dbg=0,type=-1;

    fputs("Smchsmod V1.2b\n",stdout);

    portNo = 0; opr = 'D';  /* display only */
    if (argc > 1) {
        switch (argv[1][0]) {
        case '0':
            break;
        case '1':
            portNo = 0x40;
            break;
        case '2':
            portNo = 0x80;
            break;
        case '3':
        case 'b':
        case 'B':
            portNo = 0xC0;
            break;
        default:
            usage();
        }
    }
    if (argc > 2) {
        switch (argv[2][0]) {
        case 'd':
        case 'D':
            portNo = 0;
            /* no break; */
        case 'h':
        case 'H':
        case 'l':
        case 'L':
        case 'c':
        case 'C':
            opr = (char)(toupper(argv[2][0]));
            break;
        default:
            usage();
        }
    }
    if (argc > 3) {
        type = atoi(argv[3]);
        for (n=0; idnt[n].port && n!=type; n++);
        type = n;
        printf("***Force %s***\n",idnt[type].name);
    } else
        n = 0;
#else
    char opr;
    unsigned char id,rev,r,v;
    int n,dbg=0,type=-1;

    fputs("smcdsel 1.2b\n",stdout);

    v = 0xff;   /* xxx0 0xxx: DENSEL default */
    if (argc > 1) {
        switch (argv[1][0]) {
        case 'd':
        case 'D':
            v = 0x00;   /* xxx0 0xxx: DENSEL default */
            break;
        case 'h':
        case 'H':
            v = 0x10;   /* xxx1 0xxx: DENSEL=1 */
            break;
        case 'l':
        case 'L':
            v = 0x18;   /* xxx1 1xxx: DENSEL=0 */
            break;
        default:
            usage();
        }
    }
    if (argc > 2) {
        type = atoi(argv[2]);
        for (n=0; idnt[n].port && n!=type; n++);
        type = n;
        printf("***Force %s***\n",idnt[type].name);
    } else
        n = 0;
#endif

    for (;idnt[n].port; n++) {
        _disable(); /* cli */
        /* write ini value 2 times to enter config mode */
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,idnt[n].ini);
        /* get id value */
        outp(idnt[n].port,idnt[n].idx);
        id = (unsigned char)inp(idnt[n].port+1);
        /* get rev value */
        outp(idnt[n].port,idnt[n].idx+1);
        rev = (unsigned char)inp(idnt[n].port+1);
        outp(idnt[n].port,0xaa);    /* exit config mode */
        _enable();  /* sti */
        if (opr == 'C')
            printf("*Check %s: port=%04x, ini=%02x, idx=%02x -> id=%02x, rev=%02x\n"
            ,idnt[n].name,idnt[n].port,idnt[n].ini,idnt[n].idx,id,rev);
        if (n == type)  /* force n'th */
            break;
        /* id & rev much ? */
        if ((idnt[n].id==id) && (idnt[n].rev == rev))
            break;
        /* id much only, except 665,666 */
        if ((idnt[n].id < 0x65) && (idnt[n].id == id))
            break;
    }

#ifndef DSEL
    if (idnt[n].type > 1)
        printf("SMC37C%s at %04xh\n",idnt[n].name,idnt[n].port);

    switch(idnt[n].type) {
    case 0: /* Unknown */
    case 1: /* 665GT,666GT */
    default: /* internal error.. */
        fputs("!!! Not SMC37C665IR/666IR/669/669FR/67x/68x/93x/93xAPM/93xFR/957FR chip !!!\n",stdout);
        break;
    case 2: /* 665IR,666IR */
        _disable(); /* cli */
        /* write ini value 2 times to enter config mode */
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,0x01);
        cr1 = (unsigned char) inp(idnt[n].port+1);
        outp(idnt[n].port,0x02);
        cr2 = (unsigned char) inp(idnt[n].port+1);
        /* select sr0c */
        outp(idnt[n].port,0x0c);
        r = (unsigned char) inp(idnt[n].port+1);
        if (opr == 'H')
            r |= portNo;
        if (opr == 'L')
            r &= ~portNo;
        outp(idnt[n].port+1,r);
        r = (unsigned char) inp(idnt[n].port+1);
        outp(idnt[n].port,0xaa);    /* exit config mode */
        _enable();  /* sti */
        for (v=0x40,i=1;i<=2;++i) {
            printf("Port%d",i);
            if (cr2 & 0x4) {
                switch (cr2 & 0x3) {
                case 00: portAdr[i]=0x3F8;break;
                case 01: portAdr[i]=0x2F8;break;
                case 02:
                    switch(cr1 & 0x30){
                    case 0x00: portAdr[i]=0x338;break;
                    case 0x10: portAdr[i]=0x3E8;break;
                    case 0x20: portAdr[i]=0x2E8;break;
                    case 0x30: portAdr[i]=0x220;break;
                    }
                    break;
                case 03:
                    switch(cr1 & 0x30){
                    case 0x00: portAdr[i]=0x238;break;
                    case 0x10: portAdr[i]=0x2E8;break;
                    case 0x20: portAdr[i]=0x2E0;break;
                    case 0x30: portAdr[i]=0x228;break;
                    }
                    break;
                }
                printf("(%04xh) Serial Speed Mode: %s\n",portAdr[i],v&r ?"High":"Low" );
            } else {
                printf(" disabled\n");
            }
            cr2 >>= 4;
            v <<= 1;
        }
        break;
    case 3: /* 669 */
        _disable(); /* cli */
        /* write ini value 2 times to enter config mode */
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,0x24);
        portAdr[1] =  (inp(idnt[n].port+1) & 0xFE) << 2;
        outp(idnt[n].port,0x25);
        portAdr[2] =  (inp(idnt[n].port+1) & 0xFE) << 2;
        /* select sr0c */
        outp(idnt[n].port,0x0c);
        r = (unsigned char) inp(idnt[n].port+1);
        if (opr == 'H')
            r |= portNo;
        if (opr == 'L')
            r &= ~portNo;
        outp(idnt[n].port+1,r);
        r = (unsigned char) inp(idnt[n].port+1);
        outp(idnt[n].port,0xaa);    /* exit config mode */
        _enable();  /* sti */
        for (v=0x40,i=1;i<=2;++i) {
            printf("Port%d",i);
            if (portAdr[i] & 0x300)
                printf("(%04xh) Serial Speed Mode: %s\n",portAdr[i],v&r ?"High":"Low" );
            else
                printf(" disabled\n");
            v <<= 1;
        }
        break;
    case 4: /* 67x,68x,93x,957 */
        for (i = 1; i <=2 ; ++i) {  /* Device 4=Port1, 5=Port2 */
            _disable(); /* cli */
            /* write ini value 2 times to enter config mode */
            outp(idnt[n].port,idnt[n].ini);
            outp(idnt[n].port,idnt[n].ini);
            outp(idnt[n].port,0x07);/* Select device register */
            outp(idnt[n].port+1,i+3);   /* Set device value */
            outp(idnt[n].port,0x60);
            portAdr[i] = inp(idnt[n].port+1) << 8;
            outp(idnt[n].port,0x61);
            portAdr[i] |= inp(idnt[n].port+1);
            outp(idnt[n].port,0x30);
            if (inp(idnt[n].port+1)==0)
                portAdr[i] = 0;
            outp(idnt[n].port,0xF0);/* Select serial config register */
            if ((0x20 << i) & portNo) {
                r = (unsigned char)inp(idnt[n].port+1);
                if (opr == 'H')
                    r |= 2;
                if (opr == 'L')
                    r &= ~2;
                outp(idnt[n].port+1,r);
            }
            r = (unsigned char)inp(idnt[n].port+1);
            outp(idnt[n].port,0xaa);    /* exit config mode */
            _enable();  /* sti */
            if (opr == 'C')
                printf("*Check srf0: %02x -- ",r);
            printf("Port%d",i);
            if (portAdr[i])
                printf("(%04xh) Serial Speed Mode: %s\n",portAdr[i],r&2 ?"High":"Low" );
            else
                printf(" disabled\n");
        }
        break;
    }
    if (opr == 'C')
        srdump(n,0,0x2F);
    exit(0);
}

void usage(void)
{
    fputs("usage:Smchsmod port# [h/l/d]\n",stderr);
    exit(1);
}
#else
    switch(idnt[n].type) {
    case 0: /* Unknown */
    default: /* internal error.. */
        fputs("!!! Not SMC37C665[GT|IR]/666[GT|IR]/669[FR]/67x/68x/93x[APM|FR]/957FR chip !!!\n",stdout);
        exit(1);
        break;
    case 1: /* 665GT,666GT */
    case 2: /* 665IR,666IR */
    case 3: /* 669 */
        _disable(); /* cli */
        /* write ini value 2 times to enter config mode */
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,idnt[n].ini);
        /* select sr05 */
        outp(idnt[n].port,0x05);
        r = (unsigned char) inp(idnt[n].port+1);
        if (v != 0xff)
            outp(idnt[n].port+1,(r & ~0x18) | v);
        r = (unsigned char) inp(idnt[n].port+1);
        break;
    case 4: /* 67x,68x,93x,957 */
        _disable(); /* cli */
        /* write ini value 2 times to enter config mode */
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,idnt[n].ini);
        outp(idnt[n].port,0x07);/* Select device register */
        outp(idnt[n].port+1,0); /* Set device value(FD=0) */
        outp(idnt[n].port,0xF1);/* Select FDD option register */
        r = (unsigned char) inp(idnt[n].port+1);
        if (v != 0xff)
            outp(idnt[n].port+1,(r & ~0x0c) | (v >> 1) );
        r = (unsigned char) (inp(idnt[n].port+1) << 1);
        break;
    } /* end switch */
    outp(idnt[n].port,0xaa);    /* exit config mode */
    _enable();  /* sti */
    printf("SMC37C%s at %04xh - DENSEL = ",idnt[n].name,idnt[n].port);
    switch (r & 0x18) {
    case 0x00:
        fputs("Default\n",stdout);
        break;
    case 0x18:
        fputs("Low\n",stdout);
        break;
    case 0x10:
        fputs("High\n",stdout);
        break;
    default:
        fputs("Unknown",stdout);
        break;
    }
    exit(0);
}

void usage(void)
{
    fputs("usage:smcdsel [h/l/d]\n",stderr);
    exit(1);
}
#endif

#if defined(unix)
#if defined(linux)
#define PORT_DEV "/dev/port"
#elif defined(__FreeBSD__)
#define PORT_DEV "/dev/io"
#endif /* !defined(linux) && defined(__FreeBSD__) */

FILE *fp;

void port_init(void) {
    if ((fp = fopen(PORT_DEV, "r+")) == NULL) {
        perror("error in fopen " PORT_DEV);
        exit(2);
    }
}

void port_finish(void) {
    fclose(fp);
}

#if defined(linux)

void port_error(const char *msg) {
    perror(msg);
    fclose(fp);
    exit(2);
}

int inp(unsigned port) {
    char buf[1];

#if 0
    fprintf(stderr, "inp: port = 0x%x\n", port);
#endif /* DEBUG */
    if (fseek(fp, port, SEEK_SET)) {
        port_error("error in fseek " PORT_DEV);
    }
    if (!fread(buf, 1, 1, fp)) {
        port_error("error in fread " PORT_DEV);
    }
    return buf[0];
}

void outp(unsigned port, unsigned char val) {
#if 0
    fprintf(stderr, "outp: port = 0x%x, val = 0x%x\n", port, val);
#endif /* DEBUG */
    if (fseek(fp, port, SEEK_SET)) {
        port_error("error in fseek " PORT_DEV);
    }
    if (!fwrite(&val, 1, 1, fp)) {
        port_error("error in fwrite " PORT_DEV);
    }
}

#endif /* defined(linux) */
#endif /* defined(unix) */

void srdump(int n,int start,int end)
{
    int i;
    unsigned char buf[256], *bp;

    if ((end-start) > 255) {
        fprintf(stderr,"Too big to dump\n");
        return;
    }
    printf("   +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F");

    bp = &buf[0];
    _disable(); /* cli */
    /* write ini value 2 times to enter config mode */
    outp(idnt[n].port,idnt[n].ini);
    outp(idnt[n].port,idnt[n].ini);
    for (i=start;i<=end;++i) {
        outp(idnt[n].port,i);
        *bp++ = (unsigned char)inp(idnt[n].port+1);
    }
    outp(idnt[n].port,0xaa);    /* exit config mode */
    _enable();  /* sti */
    bp = &buf[0];
    for (i=start;i<=end;++i) {
        if (((i-start)%16)==0)
            printf("\n%02x:",i);
        printf("%02x ",*bp++);
    }
    printf("\n");
}
